function results = HJB_dynamics(parameters,results, pi_n, r_bar_n, Y_n)
% This codes computes the HJB equation in a Bewley economy with nominal
% long-term debt 
% Developped by Nuo and Thomas (2015)

%PARAMETERS
delta         = parameters.delta;            % debt duration 1/lambda
gamma         = parameters.gamma;            % CRRA utility with parameter s
rho           = parameters.rho;              % discount rate
z1            = parameters.z1;               % labor productivity
z2            = parameters.z2;
la1           = parameters.la1;              % transition prob
la2           = parameters.la2;
I             = parameters.I;                % numer of points
amin          = parameters.amin;             % borrowing constraint
amax          = parameters.amax;             % max value of assets
maxit         = parameters.maxit;            % max number of iterations
crit          = parameters.crit;             % convergence ctrit 
Delta         = parameters.Delta;            % Delta in HJB
dt            = parameters.dt;               % dt in HJB
N             = parameters.N;                % Time period in dt
psi           = parameters.psi;              % inflation costs

%VARIABLES
a   = linspace(amin,amax,I)';               % assets 1D  - I elements
da  = (amax-amin)/(I-1);     
z   = [z1,z2];                              % product 1D - 2 elements
la  = [la1,la2];    
aa  = [a,a];                                % assets 2D
zz  = ones(I,1)*z;                          % product 2D
dVf = zeros(I,2);
dVb = zeros(I,2);
c   = zeros(I,2);

Aswitch = [-speye(I)*la(1),speye(I)*la(1);speye(I)*la(2),-speye(I)*la(2)]; 

Q_n = zeros(1,N);
V_n = zeros(I,2,N);
c_n = zeros(I,2,N);
A_n = cell(1,N);

%FINAL VALUE

Q_n(N)     = results.Q;
V_n(:,:,N) = results.V;
c_n(:,:,N) = results.c;
A_n{N}     = results.A;

%MAIN LOOP
for n = N-1:-1:1  
    V            = V_n(:,:,n+1);
    pi           = pi_n(n);
    Y            =Y_n(n);
    
    %bond pricing
    r_bar = r_bar_n(n);
    Q   =  (Q_n(n+1)+ dt *(delta) ) /(1 + dt*(r_bar + pi +delta)) ;
    r   =  (delta / Q -delta -pi);               % real yield
    
    % forward difference
    dVf(1:I-1,:) = (V(2:I,:)-V(1:I-1,:))/da;
    dVf(I,:)     = Q*(Y*z + r.*amax*Q).^(-gamma);    % will never be used, but impose state constraint a<=amax just in case
    % backward difference
    dVb(2:I,:)   = (V(2:I,:)-V(1:I-1,:))/da;
    dVb(1,:)     = Q*(Y*z + r.*amin*Q).^(-gamma);    % state constraint boundary condition
    
    I_concave = dVb > dVf;                         % indicator whether value function is concave (problems arise if this is not the case)
    
    %consumption and savings with forward difference
    cf           = (dVf/Q).^(-1/gamma);
    ssf          = Y*zz/Q + r.*aa - cf/Q;
    %consumption and savings with backward difference
    cb           = (dVb/Q).^(-1/gamma);
    ssb          = Y*zz/Q + r.*aa - cb/Q;
    %consumption and derivative of value function at steady state
    c0           = Y*zz + r.*aa*Q;
    dV0          = Q*c0.^(-gamma);
    
    % dV_upwind makes a choice of forward or backward differences based on
    % the sign of the drift    
    If           = ssf > 0; %positive drift --> forward difference
    Ib           = ssb < 0; %negative drift --> backward difference
    I0           = (1-If-Ib); %at steady state
    %make sure backward difference is used at amax
    %Ib(I,:) = 1; If(I,:) = 0;
    
    
    dV_Upwind    = dVf.*If + dVb.*Ib + dV0.*I0; %important to include third term
    c            = (dV_Upwind/Q).^(-1/gamma);
    u        = (c.^(1-gamma)-1)/(1-gamma) - psi/2 *pi^2;
    
    %CONSTRUCT MATRIX A
    X            = - min(ssb,0)/da;
    Y            = - max(ssf,0)/da + min(ssb,0)/da;
    Z            = max(ssf,0)/da;
    
    A1           = spdiags(Y(:,1),0,I,I) + spdiags(X(2:I,1),-1,I,I) + spdiags([0;Z(1:I-1,1)],1,I,I);
    A2           = spdiags(Y(:,2),0,I,I) + spdiags(X(2:I,2),-1,I,I) + spdiags([0;Z(1:I-1,2)],1,I,I);
    A            = [A1,sparse(I,I);sparse(I,I),A2] + Aswitch;
    B            = (rho + 1/dt)*speye(2*I) - A;
    
    u_stacked    = [u(:,1);u(:,2)];
    V_stacked    = [V(:,1);V(:,2)];
    
    b            = u_stacked + V_stacked/dt;
    V_stacked    = B\b; %SOLVE SYSTEM OF EQUATIONS
    
    V            = [V_stacked(1:I),V_stacked(I+1:2*I)]; 

    Q_n(n)     = Q;
    V_n(:,:,n) = V;
    c_n(:,:,n) = c;
    A_n{n}     = A; 
end


% RESULTS
results.Q_n  = Q_n;
results.V_n  = V_n;
results.c_n  = c_n;
results.A_n  = A_n;
